/*****************************************************************************
    memtrace.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2005 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include "params.h"
#include "globals.h"
#include "logfun.h"
#include "memtrace.h"

#ifdef MALLOC_TRACE

#define RLISTSIZ   8192
#define MAXNAMSIZ  32

#define MT_UNKNOWN 0
#define MT_MALLOC  1
#define MT_REALLOC 2
#define MT_STRDUP  3
#define MT_FREE    4
#define MT_CALLOC  5
#define MT_START   6
#define MT_END     7

static char *names[] = { "???" ,"malloc","realloc","strdup","free","calloc" };

struct list_s {	void *p1;
                void *p2;
                size_t size;
                int  ftyp;
                int  line; 
                char func[MAXNAMSIZ]; };

static struct list_s list[RLISTSIZ];
static int nentry = 0;
static int ntypes[5] = { 0,0,0,0,0 };
static int enabled = 1;
static int init = 0;
static char *lowlimit = NULL;

static int memtrace_store(int ftyp, char *func, int line, 
                           void *p1, void *p2, size_t size)
{
    if (!enabled)
        return -1;

    if (!init) {
        init = 1;
        lowlimit = malloc(1);
    }

    if (nentry < RLISTSIZ) {
        switch (ftyp) {
            case 'm':
                list[nentry].ftyp = MT_MALLOC;
                break;
            case 'c':
                list[nentry].ftyp = MT_CALLOC;
                break;
            case 'r':
                list[nentry].ftyp = MT_REALLOC;
                break;
            case 's':
                list[nentry].ftyp = MT_STRDUP;
                break;
            case 'f':
                if (p1 == NULL) {
                    LOG_MSG1 (LOG_ERR, "%s@%d: attempt to free NULL", 
                              func, line);
                    return 0;
                }
                else if ((char *)p1 <= lowlimit) {
                    LOG_MSG1 (LOG_ERR, "%s@%d: free(%p) out of bounds", 
                              func, line, p1);
                }
                list[nentry].ftyp = MT_FREE;
                break;
            case 'S':		/* start checkpoint area */
                list[nentry].ftyp = MT_START;
                break;
            case 'E':		/* end checkpoint area   */
                list[nentry].ftyp = MT_END;
                break;
            default:
                list[nentry].ftyp = MT_UNKNOWN;
                break;
        }
        list[nentry].line = line;
        list[nentry].p1   = p1;
        list[nentry].p2   = p2;
        list[nentry].size = size;
        strncpy(list[nentry].func, func, MAXNAMSIZ-1);
        list[nentry].func[MAXNAMSIZ-1] = '\0';
        ntypes[list[nentry].ftyp]++;
        return nentry++;
    }
    else
        LOG_MSG1 (LOG_ERR, "%s trace list overflow", __FUNCTION__);
        return -1;
}

void x_Free(char *func, int line, char **p)
{
    memtrace_store('f', func, line, *p, NULL, 0);
    if (*p != NULL)
        free(*p);
    *p = NULL;
}
   
void *xMalloc(char *func, int line, size_t s)
{
    void *p;

    p = malloc(s);
    if (p != NULL)
        memtrace_store('m', func, line, p, NULL, s);
    return p;
}

void *xCalloc(char *func, int line, size_t n, size_t s)
{
    void *p;

    p = calloc(n,s);
    if (p != NULL)
        memtrace_store('c', func, line, p, NULL, n*s);
    return p;
}

void *xRealloc(char *func, int line, void *o, size_t s)
{
    void *p;

    p = realloc(o,s);
    if (p != NULL)
        memtrace_store('r', func, line, p, o, s);
    return p;
}

void *xStrdup(char *func, int line, void *o)
{
    void *p;

    p = strdup(o);
    if (p != NULL)
        memtrace_store('s', func, line, p, NULL, strlen(p));
    return p;
}

void xResume(void)
{
    int i, k;

    enabled = 0;

    for (i=0; i<5; i++)
        if (ntypes[i])
            LOG_MSG5 (LOG_DEBUG, "%s(): %d", names[i], ntypes[i]);
    
    for (i=0; i<nentry; i++) {
        if (list[i].p1 == NULL) continue;
        if (list[i].ftyp == MT_FREE) {
            for (k=0; k<i; k++) {
                if (list[i].p1 == list[k].p1) {
                    if (list[k].ftyp == MT_FREE) {
                        LOG_MSG5 (LOG_DEBUG, "orphan %s[%d] %s(): "
                                  "looking for allocator of %s[%d] free()",
                                  list[k].func, list[k].line, 
                                  names[list[k].ftyp],
                                  list[i].func, list[i].line);
                        list[k].p1 = NULL;
                        list[i].p1 = NULL;
                    }
                    else {
                        LOG_MSG5 (LOG_DEBUG, "%s[%d] %s() -> %s[%d] free()",
                                  list[k].func, list[k].line, 
                                  names[list[k].ftyp],
                                  list[i].func, list[i].line);
                        list[k].p1 = NULL;
                        list[i].p1 = NULL;
                    }
                    break;
                }
            }
            if (k >= i)
                LOG_MSG5 (LOG_DEBUG, "%s[%d] free(%p) illegal", 
                          list[i].func, list[i].line, list[i].p1);
        }
    }

    for (i=0; i<nentry; i++) {
        if (list[i].p1 == NULL) continue;
        if (list[i].ftyp != MT_FREE) {
            LOG_MSG5 (LOG_DEBUG, "%s[%d] %s(%p) never free()'ed", 
                      list[i].func, list[i].line, 
                      names[list[i].ftyp], list[i].p1);
        }
    }
    nentry = 0;
    enabled = 1;
}
#endif
